/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

#include "fggl/gfx/ogl/types.hpp"
#include <fggl/gfx/ogl/shader.hpp>
#include "fggl/gfx/ogl4/fallback.hpp"

#include <iostream>
#include <vector>
#include <spdlog/spdlog.h>

namespace fggl::gfx {

	bool ShaderCache::compileShaderFromSource(const std::string &source, GLuint sid) {
		// upload and compile shader
		const char *src = source.c_str();
		glShaderSource(sid, 1, &src, nullptr);
		glCompileShader(sid);

		// check it worked
		GLint compiled = GL_FALSE;
		glGetShaderiv(sid, GL_COMPILE_STATUS, &compiled);
		if (compiled == GL_FALSE) {

			GLint maxLength = 0;
			glGetShaderiv(sid, GL_INFO_LOG_LENGTH, &maxLength);
			char *infoLog = new char[maxLength];
			glGetShaderInfoLog(sid, maxLength, &maxLength, infoLog);

			spdlog::warn("could not compile shader source: {}", infoLog);
			delete[] infoLog;
			return false;
		}

		return true;
	}

	bool ShaderCache::readAndCompileShader(const std::string &filename, GLuint shader) {
		std::string source;
		bool result = m_storage->load(fggl::data::Data, filename, &source);
		if (!result) {
			spdlog::warn("could not load shader source from disk: {}", filename.c_str());
			return false;
		}

		return compileShaderFromSource(source, shader);
	}

	ShaderCache::ShaderCache(fggl::data::Storage *storage) : m_storage(storage), m_shaders(), m_binary(true) {

		if (!GLAD_GL_ARB_get_program_binary) {
			spdlog::warn("the graphics card doesn support shader caching, disabling");
			m_binary = false;
		} else {
			// debug - disable shader cache
			m_binary = false;
		}

		if (GLAD_GL_ARB_shading_language_include) {
			setupIncludes();
		}
		initFallbackPipelines();
	}

	void ShaderCache::setupIncludes() {
		auto root = m_storage->resolvePath(data::StorageType::Data, "include");
		auto paths = m_storage->findResources(root, ".glsl");

		for (auto &path : paths) {
			std::string source;
			m_storage->load(fggl::data::Data, path.string(), &source);

			auto relPath = std::filesystem::relative(path, root);
			const auto relPathStr = "/" + relPath.string();
			glNamedStringARB(GL_SHADER_INCLUDE_ARB, -1, relPathStr.c_str(), -1, source.c_str());
		}

	}

	bool ShaderCache::loadFromDisk(GLuint pid, const std::string &pipelineName) {

		BinaryCache cache;
		auto fname = "shader_" + pipelineName + ".bin";
		bool status = m_storage->load(fggl::data::Cache, fname, &cache);

		if (!status) {
			spdlog::info("cached shader '{}' could not be loaded from disk", pipelineName);
			return false;
		}

		bool result = cacheLoad(pid, &cache);
		std::free(cache.data);
		return result;
	}

	void ShaderCache::saveToDisk(GLuint pid, const std::string &pipelineName) {
		BinaryCache cache;
		cacheSave(pid, &cache);

		auto fname = "shader_" + pipelineName + ".bin";
		m_storage->save(fggl::data::Cache, fname, &cache);
	}

	ShaderCache::ShaderPtr ShaderCache::getOrLoad(const ShaderConfig &config) {
		try {
			return m_shaders.at(config.name);
		} catch (std::out_of_range &e) {
			return load(config);
		}
	}

	ShaderCache::ShaderPtr ShaderCache::get(const std::string &name) {
		return m_shaders.at(name);
	}

	ShaderCache::ShaderPtr ShaderCache::load(const ShaderConfig &config) {
		spdlog::debug("starting shader program generation for {}", config.name);

		GLuint pid = glCreateProgram();

		if (m_binary) {
			// if we have support for shader cache, give that a go
			bool worked = loadFromDisk(pid, config.name);
			if (worked) {
				auto shader = std::make_shared<ogl::Shader>(pid);
				m_shaders[config.name] = shader;
				return shader;
			}

			spdlog::debug("could not use cached shader for '{}', doing full compile.", config.name);
		}

		// TODO actual shader loading
		GLuint vertShader = glCreateShader(GL_VERTEX_SHADER);
		readAndCompileShader(config.vertex, vertShader);
		glAttachShader(pid, vertShader);

		GLuint fragShader = glCreateShader(GL_FRAGMENT_SHADER);
		readAndCompileShader(config.fragment, fragShader);
		glAttachShader(pid, fragShader);

		GLuint geomShader = 0;
		if (config.hasGeom) {
			geomShader = glCreateShader(GL_GEOMETRY_SHADER);
			readAndCompileShader(config.geometry, geomShader);
			glAttachShader(pid, geomShader);
		}

		glLinkProgram(pid);
		glDetachShader(pid, vertShader);
		glDetachShader(pid, fragShader);

		if (config.hasGeom) {
			glDetachShader(pid, geomShader);
		}

		GLint linked = GL_FALSE;
		glGetProgramiv(pid, GL_LINK_STATUS, &linked);
		if (linked == GL_FALSE) {

			// get the error
			std::array<char, 512> infoLog;
			glGetProgramInfoLog(pid, infoLog.size(), nullptr, infoLog.data());
			spdlog::warn("linking shader program '{}' failed: {}", config.name, infoLog.data());

			// cleanup
			glDeleteProgram(pid);
			glDeleteShader(vertShader);
			glDeleteShader(fragShader);
			if (config.hasGeom) {
				glDeleteShader(geomShader);
			}

			return nullptr;
		}

		if (m_binary) {
			saveToDisk(pid, config.name);
		}

		// update the cache and return
		auto shaderPtr = std::make_shared<ogl::Shader>(pid);
		m_shaders[config.name] = shaderPtr;
		return shaderPtr;
	}

	ShaderCache::ShaderPtr ShaderCache::load(const ShaderSources &sources, bool allowBinaryCache) {

		spdlog::debug("starting shader program generation for {}", sources.name);

		GLuint pid = glCreateProgram();

		if (m_binary && allowBinaryCache) {
			// if we have support for shader cache, give that a go
			bool worked = loadFromDisk(pid, sources.name);
			if (worked) {
				auto shader = std::make_shared<ogl::Shader>(pid);
				m_shaders[sources.name] = shader;
				return shader;
			}

			spdlog::debug("could not use cached shader for '{}', doing full compile.", sources.name);
		}

		// TODO actual shader loading
		GLuint vertShader = glCreateShader(GL_VERTEX_SHADER);
		compileShaderFromSource(sources.vertexSource, vertShader);
		glAttachShader(pid, vertShader);

		GLuint fragShader = glCreateShader(GL_FRAGMENT_SHADER);
		compileShaderFromSource(sources.fragmentSource, fragShader);
		glAttachShader(pid, fragShader);

		GLuint geomShader = INVALID_SHADER_ID;
		if (!sources.geometrySource.empty()) {
			geomShader = glCreateShader(GL_GEOMETRY_SHADER);
			compileShaderFromSource(sources.geometrySource, geomShader);
			glAttachShader(pid, geomShader);
		}

		glLinkProgram(pid);
		glDetachShader(pid, vertShader);
		glDetachShader(pid, fragShader);

		if (geomShader != INVALID_SHADER_ID) {
			glDetachShader(pid, geomShader);
		}

		GLint linked = GL_FALSE;
		glGetProgramiv(pid, GL_LINK_STATUS, &linked);
		if (linked == GL_FALSE) {

			// get the error
			std::array<char, 512> infoLog{};
			glGetProgramInfoLog(pid, infoLog.size(), nullptr, infoLog.data());
			spdlog::warn("linking shader program '{}' failed: {}", sources.name, infoLog.data());

			// cleanup
			glDeleteProgram(pid);
			glDeleteShader(vertShader);
			glDeleteShader(fragShader);
			if (geomShader != INVALID_SHADER_ID) {
				glDeleteShader(geomShader);
			}

			return nullptr;
		}

		if (m_binary && allowBinaryCache) {
			saveToDisk(pid, sources.name);
		}

		// update the cache and return
		m_shaders[sources.name] = std::make_shared<ogl::Shader>(pid);
		return m_shaders[sources.name];
	}

	void ShaderCache::cacheSave(GLuint pid, BinaryCache *cache) {
		GLsizei length;
		glGetProgramiv(pid, GL_PROGRAM_BINARY_LENGTH, &length);

		cache->data = std::malloc(length);
		cache->size = length;

		glGetProgramBinary(pid, length, &cache->size, &cache->format, cache->data);
	}

	bool ShaderCache::cacheLoad(GLuint pid, const BinaryCache *cache) {
		if (!m_binary) {
			return false;
		}
		glProgramBinary(pid, cache->format, cache->data, cache->size);

		// check it loaded correctly
		GLint status = GL_FALSE;
		glGetProgramiv(pid, GL_LINK_STATUS, &status);
		return status == GL_TRUE;
	}

	void ShaderCache::initFallbackPipelines() {
		// canvas fallback pipeline
		load({
				 .name = ogl4::FALLBACK_CANVAS_PIPELINE,
				 .vertexSource = ogl4::FALLBACK_CANVAS_VERTEX_SHADER,
				 .fragmentSource = ogl4::FALLBACK_CANVAS_FRAGMENT_SHADER,
				 .geometrySource = ""
			 }, false);
	}

}

	template<>
	bool fggl::data::fggl_deserialize(std::filesystem::path &data, fggl::gfx::BinaryCache *out) {
		auto f =
			#ifdef _MSC_VER
			_wfopen(data.c_str(), L"r");
			#else
			std::fopen(data.c_str(), "r");
		#endif

		if (f == nullptr) {
			spdlog::warn("could not load cached shader, fp was null");
			return false;
		}

		auto rsize = std::fread(&out->format, sizeof(GLenum), 1, f);
		if (rsize != 1) {
			spdlog::warn("could not load cached shader: type read failed");
			std::fclose(f);
			return false;
		}

		out->size = 0;
		rsize = std::fread(&out->size, sizeof(GLsizei), 1, f);
		if (rsize != 1) {
			spdlog::warn("could not load cached shader: size read failed");
			std::fclose(f);
			return false;
		}

		out->data = std::malloc(out->size);
		auto readSize = std::fread(out->data, out->size, 1, f);

		auto result = true;
		if (readSize != 1) {
			spdlog::warn("could not load cached shader: reading failed!");
			std::free(out->data);
			result = false;
		}

		std::fclose(f);
		return result;
	}

#include <iostream>
#include <fstream>

template<>
bool fggl::data::fggl_deserialize(std::filesystem::path &data, std::string *out) {
	std::ifstream ifs(data);
	out->assign((std::istreambuf_iterator<char>(ifs)),
				(std::istreambuf_iterator<char>()));
	return true;
}

template<>
bool fggl::data::fggl_serialize(std::filesystem::path &data, const fggl::gfx::BinaryCache *out) {

	// try and write
	auto f =
		#ifdef _MSC_VER
		_wfopen( data.c_str(), L"w");
		#else
		std::fopen(data.c_str(), "w");
	#endif

	if (f == nullptr) {
		return false;
	}

	std::fwrite(&out->format, sizeof(GLenum), 1, f);
	std::fwrite(&out->size, sizeof(GLsizei), 1, f);
	std::fwrite(out->data, out->size, 1, f);

	std::fclose(f);
	std::free(out->data);

	return true;
}
