/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 24/07/22.
//

#include "fggl/gfx/ogl4/module.hpp"
#include "fggl/gfx/phong.hpp"
#include "fggl/data/procedural.hpp"

#include "fggl/assets/loader.hpp"
#include <string>

namespace fggl::gfx {

	constexpr uint32_t DEFAULT_STACKS = 16;
	constexpr uint32_t DEFAULT_SLICES = 16;
	constexpr const char *SHAPE_SPHERE{"sphere"};
	constexpr const char *SHAPE_BOX{"box"};

	static void process_shape(const YAML::Node &node, data::Mesh &mesh) {
		auto transform = data::OFFSET_NONE;

		auto offset = node["offset"].as<math::vec3>(math::VEC3_ZERO);
		transform = glm::translate(transform, offset);

		auto scale = node["scale"].as<math::vec3>(math::VEC3_ONES);
		transform = glm::scale(transform, scale);
		debug::debug("scale: {}, {}, {}", scale.x, scale.y, scale.z);

		// now the shape itself
		auto type = node["type"].as<std::string>();
		if (type == SHAPE_BOX) {
			data::make_cube(mesh, transform);
		} else if (type == SHAPE_SPHERE) {
			auto stacks = node["stacks"].as<uint32_t>(DEFAULT_STACKS);
			auto slices = node["slices"].as<uint32_t>(DEFAULT_SLICES);
			data::make_sphere(mesh, transform, stacks, slices);
		} else {
			debug::log(debug::Level::warning, "unknown shape type requested: {}", type);
		}
	}

	void attach_mesh(const entity::ComponentSpec &spec, entity::EntityManager &manager, const entity::EntityID &id, modules::Services &services) {

		// check for the asset service
		auto* assetService = services.get<assets::AssetManager>();
		auto* assetLoader = services.get<assets::Loader>();
		if ( assetService == nullptr || assetLoader == nullptr ) {
			// no asset service, give up
			return;
		}

		// asset is a procedural mesh description
		if (spec.has("shape")) {
			data::Mesh* meshAsset = nullptr;
			auto pipeline = spec.get<std::string>("pipeline", "");

			// check if we had previously loaded this asset
			const auto shapeName = spec.get<std::string>("shape_id", "");
			if ( !shapeName.empty() ) {
				meshAsset = assetService->get<data::Mesh>(shapeName);
			}

			// we've not loaded this before - generate mesh
			if ( meshAsset == nullptr ) {
				// procedural meshes are build as static meshes first
				auto* meshTmp = new data::Mesh();

				if (spec.config["shape"].IsSequence()) {
					for (const auto &node : spec.config["shape"]) {
						process_shape(node, *meshTmp);
					}
				} else {
					process_shape(spec.config["shape"], *meshTmp);
				}
				meshTmp->removeDups();

				if (!shapeName.empty()) {
					meshAsset = assetService->set(shapeName, meshTmp);
				} else {
					meshAsset = meshTmp;
				}
			}

			// TODO we need to trigger an upload to the GPU (somehow)
			#ifdef FGGL_ALLOW_DEFERRED_UPLOAD
				// the graphics stack can detect static meshes without a rendering proxy at runtime and fix it but this
				// requires loading the whole model into the ECS (and should be removed in the future). instead we should
				// be triggering the upload at this point (if needed).
				auto &entityMesh = manager.add<data::StaticMesh>(id);
				entityMesh.pipeline = pipeline;
				entityMesh.mesh = *meshAsset;
				debug::warning("HACKY: Triggered proc mesh - using deferred upload");
			#endif

			return;
		}

		// asset is a model from the resource cache
		if ( spec.has("model") ) {
			// figure out what model we want
			auto assetStr = spec.get<std::string>("model", "");
			auto* asset = assetService->get<ogl4::StaticModel>(assetStr);

			if ( asset == nullptr ) {
				// the asset is not loaded/does not exist
				debug::error("requested model {} but it was not loaded.", assetStr);
				return;
			}

			// copy the asset to the model
			auto& model = manager.add<ogl4::StaticModel>(id);
			model = *asset;
		}
	}

	void attach_material(const entity::ComponentSpec &spec,
						 entity::EntityManager &manager,
						 const entity::EntityID &id,
						 modules::Services& services ) {
		auto &mat = manager.add<gfx::PhongMaterial>(id);
		mat.ambient = spec.get<math::vec3>("ambient", gfx::DEFAULT_AMBIENT);
		mat.diffuse = spec.get<math::vec3>("diffuse", gfx::DEFAULT_DIFFUSE);
		mat.specular = spec.get<math::vec3>("ambient", gfx::DEFAULT_SPECULAR);
		mat.shininess = spec.get<float>("ambient", gfx::DEFAULT_SHININESS);
	}

	void attach_light(const entity::ComponentSpec &spec, entity::EntityManager &manager, const entity::EntityID &id, modules::Services& services) {
		auto &light = manager.add<gfx::Light>(id);
	}

	bool OpenGL4::factory(modules::ModuleService service, modules::Services &services) {
		if (service == WindowGraphics::service) {
			// setup the thing responsible for graphics
			auto *storage = services.get<data::Storage>();
			auto *fontLibrary = services.get<gui::FontLibrary>();
			auto *assets = services.get<assets::AssetManager>();

			services.bind<WindowGraphics, ogl4::WindowGraphics>(storage, fontLibrary, assets);

			// register as responsible for creating rendering components
			auto *entityFactory = services.get<entity::EntityFactory>();
			entityFactory->bind(data::StaticMesh::guid, attach_mesh);
			entityFactory->bind(gfx::PhongMaterial::guid, attach_material);
			entityFactory->bind(gfx::Light::guid, attach_light);

			return true;
		}
		return false;
	}

} // namespace fggl::gfX::ogl4