/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 20/11/2021.
//

#include <fggl/input/input.hpp>

#include <fggl/gfx/camera.hpp>
#include <fggl/input/camera_input.hpp>

namespace fggl::input {

	void process_arcball(entity::EntityManager &ecs, const Input &input, entity::EntityID cam) {
		// see https://asliceofrendering.com/camera/2019/11/30/ArcballCamera/
		auto &camTransform = ecs.get<fggl::math::Transform>(cam);
		auto &camComp = ecs.get<fggl::gfx::Camera>(cam);
		auto &mouse = input.mouse;

		glm::vec4 position(camTransform.origin(), 1.0f);
		glm::vec4 pivot(camComp.target, 1.0f);
		glm::mat4 view = glm::lookAt(camTransform.origin(), camComp.target, camTransform.up());
		glm::vec3 viewDir = -glm::transpose(view)[2];
		glm::vec3 rightDir = glm::transpose(view)[0];

		float deltaAngleX = (2 * M_PI);
		float deltaAngleY = (M_PI);
		float xAngle = (-mouse.axisDelta(fggl::input::MouseAxis::X)) * deltaAngleX;
		float yAngle = (-mouse.axisDelta(fggl::input::MouseAxis::Y)) * deltaAngleY;

		// rotate the camera around the pivot on the first axis
		glm::mat4x4 rotationMatrixX(1.0f);
		rotationMatrixX = glm::rotate(rotationMatrixX, xAngle, fggl::math::UP);
		position = (rotationMatrixX * (position - pivot)) + pivot;

		// rotate the camera aroud the pivot on the second axis
		glm::mat4x4 rotationMatrixY(1.0f);
		rotationMatrixY = glm::rotate(rotationMatrixY, yAngle, rightDir);
		glm::vec3 finalPos = (rotationMatrixY * (position - pivot)) + pivot;

		camTransform.origin(finalPos);
	}

	void process_scroll(entity::EntityManager &ecs,
						const Input &input,
						entity::EntityID cam,
						float minZoom,
						float maxZoom) {
		auto &camTransform = ecs.get<fggl::math::Transform>(cam);
		auto &camComp = ecs.get<fggl::gfx::Camera>(cam);

		const glm::vec3 dir = (camTransform.origin() - camComp.target);
		const glm::vec3 forward = glm::normalize(dir);

		glm::vec3 motion(0.0F);
		float delta = input.mouse.axis(fggl::input::MouseAxis::SCROLL_Y);
		if ((glm::length(dir) < maxZoom && delta < 0.0f) || (glm::length(dir) > minZoom && delta > 0.0f)) {
			motion -= (forward * delta);
			camTransform.origin(camTransform.origin() + motion);
		}
	}

	void process_freecam(entity::EntityManager &ecs, const Input &input, entity::EntityID cam) {
		float rotationValue = 0.0f;
		glm::vec3 translation(0.0f);

		auto &keyboard = input.keyboard;
		auto &settings = ecs.get<FreeCamKeys>(cam);

		// calculate rotation (user input)
		if (keyboard.down(settings.rotate_cw)) {
			rotationValue = ROT_SPEED;
		} else if (keyboard.down(settings.rotate_ccw)) {
			rotationValue = -ROT_SPEED;
		}

		// calculate movement (user input)
		if (keyboard.down(settings.forward)) {
			translation -= fggl::math::RIGHT;
		}

		if (keyboard.down(settings.backward)) {
			translation += fggl::math::RIGHT;
		}

		if (keyboard.down(settings.right)) {
			translation += fggl::math::FORWARD;
		}

		if (keyboard.down(settings.left)) {
			translation -= fggl::math::FORWARD;
		}

		// apply rotation/movement
		auto camTransform = ecs.get<fggl::math::Transform>(cam);
		auto camComp = ecs.get<fggl::gfx::Camera>(cam);

		glm::vec4 position(camTransform.origin(), 1.0f);
		glm::vec4 pivot(camComp.target, 1.0f);

		// apply movement
		if (translation != glm::vec3(0.0f)) {
			const auto rotation = (position - pivot);
			const float angle = atan2f(rotation.x, rotation.z);
			const auto rotationMat = glm::rotate(MAT_IDENTITY, angle, fggl::math::UP);

			auto deltaMove = (rotationMat * glm::vec4(translation, 1.0f)) * PAN_SPEED;
			deltaMove.w = 0.0f;

			position += deltaMove;
			pivot += deltaMove;
		}

		// apply rotation
		if (rotationValue != 0.0f) {
			glm::mat4 rotation = glm::rotate(MAT_IDENTITY, rotationValue, fggl::math::UP);
			position = (rotation * (position - pivot)) + pivot;
		}

		camTransform.origin(position);
		camComp.target = pivot;
	}

	void process_edgescroll(entity::EntityManager &ecs, const Input &input, entity::EntityID cam) {
		glm::vec3 translation(0.0f);

		auto &mouse = input.mouse;

		// calculate movement (user input)
		if (mouse.axis(MouseAxis::Y) < 0.9f) {
			translation -= fggl::math::RIGHT;
		}

		if (mouse.axis(MouseAxis::Y) > -0.9f) {
			translation += fggl::math::RIGHT;
		}

		if (mouse.axis(MouseAxis::X) > -0.9f) {
			translation += fggl::math::FORWARD;
		}

		if (mouse.axis(MouseAxis::X) < 0.9f) {
			translation -= fggl::math::FORWARD;
		}

		// apply rotation/movement
		auto &camTransform = ecs.get<fggl::math::Transform>(cam);
		auto &camComp = ecs.get<fggl::gfx::Camera>(cam);

		glm::vec4 position(camTransform.origin(), 1.0f);
		glm::vec4 pivot(camComp.target, 1.0f);

		// apply movement
		if (translation != glm::vec3(0.0f)) {
			const auto rotation = (position - pivot);
			const float angle = atan2f(rotation.x, rotation.z);
			const auto rotationMat = glm::rotate(MAT_IDENTITY, angle, fggl::math::UP);

			auto deltaMove = (rotationMat * glm::vec4(translation, 1.0f)) * PAN_SPEED;
			deltaMove.w = 0.0f;

			position += deltaMove;
			pivot += deltaMove;
		}

		// move camera
		camTransform.origin(position);
		camComp.target = pivot;
	}
}
