/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 03/07/22.
//

#ifndef FGGL_ASSETS_LOADER_HPP
#define FGGL_ASSETS_LOADER_HPP

#include <map>
#include <string>
#include <memory>
#include <functional>
#include <queue>
#include <variant>

#include "fggl/assets/types.hpp"
#include "fggl/data/storage.hpp"

namespace fggl::assets {

	enum class LoadType {
		DIRECT, // given pointer to persistent memory
		STAGED, // given pointer to temp memory
		PATH // given filesystem::path
	};

	struct ResourceRequest {
		AssetGUID m_guid;
		AssetType m_type;
	};

	class Loader {
		public:
			constexpr const static modules::ModuleService service = modules::make_service("fggl::assets::Loader");

			explicit inline Loader(data::Storage *storage) : m_storage(storage) {}

			explicit Loader(Loader *parent, data::Storage *storage) : m_parent(parent), m_storage(storage) {};

			// no move, no copy.
			Loader(const Loader &) = delete;
			Loader &operator=(const Loader &) = delete;
			Loader(Loader &&) = delete;
			Loader &operator=(Loader &&) = delete;

			inline void setFactory(AssetType type, Checkin fn, LoadType loading = LoadType::DIRECT) {
				m_factories[type] = std::make_pair(fn, loading);
			}

			inline void unsetFactory(AssetType type) {
				m_factories.erase(type);
			}

			inline void request(const AssetGUID &guid, const AssetType &type) {
				m_requests.push(ResourceRequest{guid, type});
			}

			void load(const AssetGUID &guid, const AssetType &type) {
				auto path = m_storage->resolvePath(data::StorageType::Data, guid);

				auto &config = m_factories.at(type);
				switch (config.second) {
				case LoadType::DIRECT:
					// TODO we load the data into main memory and give a pointer to it.
					debug::log(debug::Level::error, "Tried to load direct asset - no one wrote that yet!");
					break;
				case LoadType::STAGED:
					// TODO we load the data into temp memory and give a pointer to it.
					debug::log(debug::Level::error, "Tried to load staged asset - no one wrote that yet!");
					break;
				case LoadType::PATH: config.first(this, guid, AssetData(&path));
					break;
				}
			}

			void progress() {
				if (m_requests.empty()) {
					return;
				}
				auto &request = m_requests.front();
				load(request.m_guid, request.m_type);
				m_requests.pop();
			}

			[[nodiscard]]
			inline bool done() const {
				return m_requests.empty();
			}

			/**
			 * Complete all remaining loading requests, blocking until complete.
			 */
			inline void finish() {
				while (!done()) {
					progress();
				}
			}

		private:
			Loader *m_parent = nullptr;
			using Config = std::pair<Checkin, LoadType>;
			data::Storage *m_storage;
			std::queue<ResourceRequest> m_requests;
			std::map<AssetType, Config> m_factories;
	};

} // namespace fggl::assets

#endif //FGGL_ASSETS_LOADER_HPP
